"""
Nonworking experiement with naive Bayes classification NLP

Reference:
    Bird, Klein, Loper, 2009 "Natural Language Processing with Python", O'REILLY
"""

from collections import defaultdict, Counter

from pug.nlp.util import get_words

# count[0] = 1-gram counts
# count[1] = 2-gram counts
count = defaultdict(int)
#count.update(fish)


fish_count = Counter({'carp': 10, 'eel': 1, 'perch': 3, 'trout': 1, 'salmon': 1, 'perch': 3, 'whitefish': 2})

#count = [Counter(get_words(s))]

class Classifier:
    def __init__(self, get_features=get_words, s='', path=''):
        self.get_features = get_features
        # TODO: use a defaultdict everywhere a dict is used
        self.feature_count = Counter()
        self.category_count = defaultdict(Counter)  # dict of Counters, count of documents binned in each category 
        self.num_items = 0

    def increment_feature(self, feature, category):
        "Increase the count for a feature<->category association"
        self.feature_count[feature] += Counter((feature,))
        # self.feature_count.setdefault(feature, Counter())
        # self.feature_count[feature].setdefault(category, 0)
        # self.feature_count[feature][category] += 1

    def increment_count(self, category):
        "Increase the count of documents associated with a category"
        self.category_count.setdefault(category, 0)
        self.category_count[category] += 1

    def count_feature_in_category(self, feature, category):
        "Count of feature occurences in a category"
        # using a defaultdict everywhere would take care of this
        if feature in self.feature_count and category in self.feature_count(feature):
            return float(self.feature_count[feature][category])
        return 0.

    def items_in_category(self, category):
        if category in self.category_count:
            return float(self.category_count[category])    
        return 0.

    def num_items(self):
        "Total of the number of items that have been classified"
        # inefficient?, so cache this, if possible to detect with a dict has changed efficiently
        return sum(self.category_count.values())

    def categories(self):
        return list(self.category_count)

    def train(self, item, category):
        features = self.get_features(item)
        self.category_count[category] += Counter(features)


N_len = max(c for w, c in count.iteritems())
N = [0] * N_len + 1
for w, c in count.iteritems():
    N[c] += 1
N_total = sum(N)

V = len(set(k for k in count if isinstance(k, basestring)))

# stupid backoff method, needs example code

# from Jurafsky, Smoothing: Add-One lecture, 2:00/06:30
def mle_estimate(word, prior_word):
    """given a prior word, estimate the probability of the next (current) word"""
    float(count[(prior_word, word)]) / count[(prior_word,)]


# "Errata" in Jurafsky, Smoothing: Add-One lecture, 2:00/06:30, doesn't mention what V means (vocabulary size)
def laplace_estimate(word, prior_word, k=1, V=V):
    float(count[(prior_word, word)] + k) / (count[(prior_word,)] + k * V)


def interpolation_estimate(word, prior_word, m=1, V=V):
    float(count[(prior_word, word)] + m * (1./V)) / (count[(prior_word,)] + m)


def unigram_prior(word, prior_word, m=1, V=V):
    float(count[(prior_word, word)] + m * count(prior_word) / float(V)) / (count[prior_word] + m)

c_star = defaultdict(lambda: N[1] / N_total)

for w, c in count.iteritems():
    c_star[w] = (count(w) + 1) * N[c + 1] / N[c]


# word should be ngram and they should all be tuples, so that backoff can just remove the first and take the remainder (or nothing)
def good_turing(word, prior_word, m=1):
    c_star[word]
    





"""Dimension reduction and binary classification of nongaussian data

Gradually increases the number of dimensions until the "score" captures most of the information content, i.e.,
the mean squared error between the force-directed graph distance matrix (p_matrix) and the true position matrix
(observed vectors) is below some arbitrary threshold or ceases to be reduced significantly by adding more dimensions.
Test and training data records are classified in 2 categories with score proportional to likelihood of being in either category
(+1 for one category -1 for the other) and test/training data

NOT PEP-8 compliant (e.g. 2-space indentation instead of 4)

By Lizhong Zheng <lizhong@mit.edu> and Hobson Lane <hobson@totalgood.com>

"""
import numpy
import matplotlib.pyplot as pyplot

from numpy import zeros, ones
from numpy.random import rand


def projection_vector(p_matrix, labels):
  """Returns a projection vector for binary decision. p_matrix is a n_training x n_dimensions
  numpy array with the positions achieved from the spring function. labels is a n_training x 1 numpy array
  of +/-1 labels.
  """
  assert p_matrix.shape[0] == labels.shape[0]
  temp = numpy.copy(p_matrix)
  for idx in range(p_matrix.shape[0]):
    temp[idx][:] = labels[idx]*temp[idx][:]
  unnormalized_pv = temp.sum(axis = 0)
  return unnormalized_pv/pow(numpy.dot(unnormalized_pv, unnormalized_pv.T), 0.5)
  
def spring(d_matrix, drift, labels, n_dimensions=5, n_iteration=555):
  """Propagate the p_matrix (positions of the nodes) for n_iterations printing out the MSE matrix at the end.

    Returns the final position matrix after the force-directed graph settles out.
    d_matrix is a n_training x n_training numpy array with all the pairwise distances.
    Drift is a function that uses labels to vary the drift between each iteration.
    Which direction is steepest in your training data average of blue points and red points,
    the vectors between the groups to say which group the node is closer too
      
    Returns:
      p_matrix (np.matrix of float): each row is a node position?  the eigenvectors of the reduced dimensional space?
    Arguments:
      d_matrix (np.matrix of float): symmetric N x N matrix of distances between nodes (documents)
        column is node called `acted` the one being pulled, the column the node doing the push/pulling `actor`
      drift (bool): whether there should be some supervisory force to keep the absolute position of nodes near zero
      labels (list of int): +/- 1 for each of N nodes or 0 if no labeled. +/-1 for only 2 categories or labels in your training set
      n_dimensions (int): number of dimensions to reduce to
        We don't know how many dimensions our final vector will be, so adjust this from low to high until the gain in mse is small
      n_iterations (int): number of times
  """
  # Make sure distance matrix is square.
  assert d_matrix.shape[0] == d_matrix.shape[1]
  # # HL: FIXME: delete this unused variable
  # n_elements = d_matrix.shape[0]
  p_matrix = init_position(d_matrix, n_dimensions)

  for idx in range(n_iteration):
    # Perform drift based on the function provided.
    # Perform spring based iterative positioning. We hope this will converge
    # to our desired distance matrix.
    p_matrix = p_matrix + drift(labels, n_dimensions)
    p_matrix = propagate(p_matrix, d_matrix)
  
  # Accumulate the squared error for every pairwise distance. 
  # propogate all the forces to all the nodes for one time step
  mse = [0, 0]
  for actor in range(d_matrix.shape[0]):
    for acted in range(d_matrix.shape[0]):
      if actor != acted:
        #import pdb
        #pdb.set_trace()
        #           
        # LZ:          spring rest-length      actual distance(N-dim vector, N-dim vect)        
        mse[0] += pow(d_matrix[actor][acted]-distance(p_matrix[actor][:], p_matrix[acted][:]), 2)
        mse[1] += 1
  print mse
  return p_matrix
  

# Randomly initialize position.
def init_position(d_matrix, n_dimensions):
  n_elements = d_matrix.shape[0]
  return rand(n_elements, n_dimensions)
  
def drift(labels, n_dimensions):
  p_change = 0.01*ones(shape = (labels.shape[0], n_dimensions))
  for idx in range(p_change.shape[0]):
    if labels[idx] < 0:
      p_change[idx][:] = -p_change[idx][:]
  return p_change


def propagate(p_matrix, d_matrix):
  spring_constant = 0.1
  f_matrix = zeros(shape = p_matrix.shape)
  for acted in range(d_matrix.shape[0]):
    for actor in range(d_matrix.shape[0]):
      if acted != actor:
        #import pdb
        #pdb.set_trace()
        # LZ: force vector in N-dim                     desired distance(N-dim vector, N-dim vect)    actual distance
        f_matrix[acted][:] += -spring_constant*(distance(p_matrix[actor][:], p_matrix[acted][:])-d_matrix[actor][acted])*(p_matrix[acted][:]-p_matrix[actor][:])/distance(p_matrix[actor][:], p_matrix[acted][:])
        
  return p_matrix + f_matrix

  
# Implements dot product.
def distance(x, y):
  temp = (x - y).T
  return pow(numpy.dot(temp.T, temp), 0.5)

def outcome_map1(outcome):
  if outcome == 'PART':
    return 1
  else:
    return -1


def run_test():
  """Attempt to find the optimal clusters and dimensions for a set of documents
  To create a test distance matrix with say 100 documents just:
    1. create 100 1000-D points (high dimension) that are Gaussian iid
    2. move half (50) of the points some significant distance from the others to create 2 groups/clusters
    3. calculate the distance between all the documents to populate the 100x100 distance matrix
  * Run this distance matrix through the algorithm (propagate) as if it was the Natural language distance
  * The data should stop improving the mse at about 12 dimensions (for LiZhong's tests)
  
  Note:
    * It may not work well on our low dimensional, low information-content (short with lots of repetition of phrases) "documents"
  """
  #args = sys.argv[1:]
  #pickle_name = args[0]
  #with open(pickle_name) as f:
  #  database, diagnosis, outcome = pickle.load(f)
  #
  #for idx in range(len(outcome)):
  #  outcome[idx] = outcome_map1(outcome[idx])
  #with open("result.txt", 'w') as f:
  #  for database_idx in range(len(database)):
  #    counter_T = 0
  #    counter_F = 0
  #    for diagnosis_idx in range(len(diagnosis)):
  #      if diagnosis[diagnosis_idx] == database_idx:
  #        if outcome[diagnosis_idx] == 1:
  #          counter_T += 1
  #        else:
  #          counter_F += 1
  #    f.write(str(database_idx) + ' ' + database[database_idx] + ',    ' + str(counter_T) + ',' + str(counter_F) + '\n')
  
  
  # HL: random 6x6 matrix
  temp = 2*rand(6, 6)
  # HL: random labels +/- 1
  labels = 2*(numpy.random.randint(0, 2, (6, 1))-0.5)
  # HL: this is the "Main" work that shold take a while on a big matrix
  p_matrix = spring(temp.T+temp, drift, labels, 5, 555)
  

  
  #temp = numpy.array([[0, 0.1, 1, 1.1], [0, 0, 1.2, 1.4], [0, 0, 0, 0.05], [0, 0, 0, 0]])
  #labels = numpy.array([[1], [1], [-1], [-1]])
  #p_matrix = spring(temp.T+temp, drift, numpy.array([[1], [1], [-1], [-1]]), 2, 555)

  # Center the positions.
  p_mean = p_matrix.mean(0)
  for idx in range(p_matrix.shape[0]):
    p_matrix[idx][:] = p_matrix[idx][:]-p_mean
  
  # Find the projection vector and plot the training data.
  a = projection_vector(p_matrix, labels)
  pyplot.figure()
  for idx in range(p_matrix.shape[0]):
    if labels[idx] >= 0:
      pyplot.scatter(p_matrix[idx][0], p_matrix[idx][1], s=100, c='r')
    else:
      pyplot.scatter(p_matrix[idx][0], p_matrix[idx][1], s=100, c='b')
  ax = pyplot.axes()
  ax.arrow(0, 0, a[0], a[1], head_width=0.05, head_length=0.1, fc='k', ec='k')
  pyplot.axis([-3, 3, -3, 3])
  pyplot.show()
  
if __name__ == '__main__':
  run_test()



import mdp
import numpy as np
import scipy
import inaugural as data

occurrences, row_labels, col_labels = data.get_adjacency_matrix(sorted(data.DEFAULT_FILE_LIST), entropy_threshold=0.1, normalized=False, verbosity=1)
def reduce(occurrences=occurrences, dim=2):
    pca = mdp.nodes.PCANode(output_dim=dim, dtype='float64')
A = scipy.sparse.csr_matrix(A)
A.size
len(A.A)
len(A.A[0])
A.dtype
A = scipy.sparse.csr_matrix(A, dtype='float64')
A.A.nbytes
2500 * 56 * 8
pcanode80d = mdp.nodes.PCANode(output_dim=.8, dtype='float64')
pcanode80d.train(A)
pcanode80d.train(A.A)
pcanode80d.stop_training()
pcanode80d.avg
pca = pcanode80d
del(pcanode80d)
pca.d
pca.explained_variance
pca.get_current_train_phase
pca.get_current_train_phase()
pca.output_dim
pca.input_dim
pca.is_invertible
pca.is_invertible()
pca.has_multiple_training_phases
pca.has_multiple_training_phases()
pca.save
pca.save('pca.pickled')
pca.svd
pca.v
pca.var_abs
pca.var_part
pca.var_rel
pca.tlen
# pca.svg??
pca.svg
pca.svd
# pca.svd??
len(pca.v)
len(words_
)
len(words
)
words[pca.v.index(max(pca.v))]
max(pca.v)
words[np.where(pca.v==max(pca.v))]
words[np.where(pca.v==max(pca.v))[0]]
words
for var in pca.v:
    if var[0] > .8:
        print var[0], words[i]
for i, var in enumerate(pca.v):
    if var[0] > .8:
        print var[0], words[i]
for i, var in enumerate(pca.v):
    if var[0] > .7:
        print var[0], words[i]
for i, var in enumerate(pca.v):
    if var[0] > .5:
        print var[0], words[i]
for i, var in enumerate(pca.v):
    if var[0] > .3:
        print var[0], words[i]
for i, var in enumerate(pca.v):
    if var[0] > .2:
        print var[0], words[i]
for i, var in enumerate(pca.v):
    if var[0] > .1:
        print var[0], words[i]
for i, var in enumerate(pca.v):
    if var[0] > .05:
        print var[0], words[i]
word_var = []
for i, var in enumerate(pca.v):
    word_var += [(var[0], words[i])]
sorted(word_var)[:50]
sorted(word_var, reverse=True)[:50]
history



from scrapy.selector import Selector
from pug.nlp.util import make_int, make_float, make_us_postal_code

def get_tables(html, verbosity=0):
    """Extract a list of tables of python objects from HTML

    return a list of lists (tables) of lists (rows) of objects (cell values)

    Convert each text element to the most complete python object possible.
    >>> get_tables('<p>first table<table><thead><tr><th>H1</th><th>H2</th></tr></thead><tr><td>1</td><td>one</td></tr><tr><td>2</td><td>two</td></tr><tr><td>3</td><td>three</td></tr></table><p>malformed table<table><tr><tr><td>1</td><td>one</td></tr><tr><td>2</td><td>two</td></tr><tr><td>3</td><td>three</td></tr></table><p>empty table<table></table>')
    [[[u'H1', u'H2'], [1, u'one'], [2, u'two'], [3, u'three']],
    [[], [1, u'one'], [2, u'two'], [3, u'three']],
    []]
    """  
    dom = Selector(text=html)
    tables = dom.css('table')

    ans = []
    for table in tables:
        if verbosity > 2:
            print 'table={0}'.format(table)
        rows = table.css('tr')
        ans += [[]]
        for i, row in enumerate(rows):
            if verbosity > 2:
                print 'row {1}={0}'.format(row, i)
            values = row.css('th') or row.css('td')
            ans[-1] += [[]]
            for j, value in enumerate(values):
                s = value.css('::text').extract()
                if s:
                    s = s[0]
                if verbosity > 1:
                    print 'value({1},{2})={0}'.format(s, i, j)
                try:
                    s = float(s)
                except:
                    pass
                if isinstance(s, float):
                    try:
                        if int(s) == s:
                            s = int(s)
                    except:
                        pass
                if verbosity > 1:
                    print 'value({1},{2})={0}'.format(s, i, j)
                ans[-1][-1] += [s]
    return ans


import pug.nlp.inaugural as nlp
from tabulate import tabulate
import numpy as np

# duplicates results of tutorial at http://webhome.cs.uvic.ca/~thomo/svd.pdf

# changed died to die and new-hampshire to hampshire to avoid stemming
data = [
    'Romeo and Juliet.',
    'Juliet: O happy dagger!',
    'Romeo die by dagger.',
    '"Live free or die”, that’s the Hampshire’s motto.',
    'Did you know, New-Hampshire is in New-England.',
    ]


O, row_labels, col_labels = nlp.get_occurrence_matrix(data)
O = np.matrix(O).transpose().tolist()
row_labels, col_labels = col_labels, row_labels

keywords = ('romeo', 'juliet', 'happy', 'dagger', 'live', 'die', 'free', 'hampshire')
O_small = [O[row_labels.index(kw)] for kw in keywords if kw in row_labels]

print '\n'.join(keywords)
print tabulate(O_small, [str(cl) for cl in col_labels])

U, s, V = np.linalg.svd(O_small, full_matrices=False)
Sigma = np.diag(s)

np.allclose(O_small, np.dot(U,np.dot(Sigma, V)))

# reduced number of dimensions
k = 2
S_k = np.diag(s[:k])
U_k = np.array([U[i][:k] for i in range(U.shape[0])])
V_k = np.array([V[i] for i in range(k)])

O_k = np.dot(U_k, np.dot(S_k, V_k))

print keywords
print tabulate(S_k)
print tabulate(U_k)
print tabulate(V_k)


# decision tree for list of lists (table) classification
# attempts to identify cases that result in a service call

from examples import tobes_data
from nlp.db.explore import count_unique


#from  django.db.models import Manager
#from django.db.models.query import QuerySet

def divide(table, field, target=0, default=0, return_type=list, ignore_fields=['id', 'pk']):
    judge = None
    # TODO: DRY this up! Could make judges a dict with (type, type) as keys. See git branch 'unlambda'

    # row is a dict (potentially a QuerySet.values() object)
    keys = None
    if hasattr(table[0], 'get'):
        # convert the integer index into a dict key
        if isinstance(field, int):
            # need to sort this list of keys to get a consistent order
            keys = sorted(tuple(k for k in table[0] if k not in ignore_fields))
            keyed_field = keys[field]
        else:
            keyed_field = field
        if isinstance(target, (int, float, bool)):
            judge = lambda row: row.get(keyed_field, default) >= target
        elif isinstance(target, (tuple, list, dict, set)):
            judge = lambda row: row.get(keyed_field, default) in target
        else:
            judge = lambda row: row.get(keyed_field, default) == target
    elif isinstance(table, (list, tuple)) or isinstance(field, (int, float)):
        if isinstance(target, (int, float, bool)):
            judge = lambda row: row[field] >= target
        elif isinstance(target, (tuple, list, dict, set)):
            judge = lambda row: row[field] in target
        else:
            judge = lambda row: row[field] == target
    else:
        if isinstance(target, (int, float, bool)):
            judge = lambda row: getattr(row, field, default) >= target
        elif isinstance(target, (tuple, list, dict, set)):
            judge = lambda row: getattr(row, field, default) in target
        else:
            judge = lambda row: getattr(row, field, default) == target

    # create an iterator/generator with yield or just annotate the row
    if keys:
        true_group = return_type(return_type(row[k] for k in keys) for row in table if judge(row))
        false_group = return_type(return_type(row[k] for k in keys) for row in table if not judge(row))
    else:
        true_group = return_type(row for row in table if judge(row))
        false_group = return_type(row for row in table if not judge(row))

    # import inspect
    # print inspect.getsource(judge)
    
    return (true_group, false_group)


def gini_impurity(table, field=-1):
    """Gini impurity evaluation of predictions

    Returns the probability [0, 1], that the wrong category/prediction has been assigned.
    """
    try:
        N = table.count()
    except:
        N = len(table)
    counts = count_unique(table, field)
    impurity = 0.0
    for k1 in counts:
        p1 = float(counts[k1]) / N
        for k2 in counts:
            if not k1 == k2:
                p2 = float(counts[k2]) / N
                impurity += p1 * p2
    return impurity


def entropy(table, field=-1, num_categories=2):
    """Total entropy for all the categorizations assigned

    sum(p(x) * log(p(x)) for x in count_unique(table, field)

    Which measures how different each categorization is from the others
    """
    from math import log
    counts = count_unique(table, field)
    entropy = 0.0
    try:
        N = table.count()
    except:
        N = len(table)
    for k in counts:
        p = float(counts[k]) / N
        entropy -=  p * log(p, num_categories)
    return entropy


def entropy_and_impurity(table, field=-1, num_categories=2):
    """Gini impurity evaluation of predictions

    Returns the probability [0, 1], that the wrong category/prediction has been assigned.

    >>> entropy_and_impurity(tobes_data, -1)  # doctest: +ELLIPSIS
    (1.50524..., 0.6328125)
    """
    from math import log
    try:
        N = table.count()
    except:
        N = len(table)
    counts = count_unique(table, field)
    impurity = 0.0
    entropy = 0.0
    for k1 in counts:
        p1 = float(counts[k1]) / N
        entropy -= p1 * log(p1, num_categories)
        for k2 in counts:
            if not k1 == k2:
                p2 = float(counts[k2]) / N
                impurity += p1 * p2
    return entropy, impurity
# TODO: add class attributes `num_categories` and `field`
#       to allow configuration of fun for use in a map, lambda, etc


def impure_entropy(table, field=-1):
    e, i = entropy_and_impurity(table, field)
    return e * i


# def build_tree(table, field, scorer=entropy):
#     if not table:
#         return decision_node()
#     score = scorer(table, field)

#     best_gain = 0.0
#     best_sets = None

#     column_values = {}
#     M = len(table[0])
#     for row in table:
#         for i in range(M):
#             column_values[get(row, i)] = 1
#     pass


class DecisionNode(object):
    def __init__(self, col=-1, value=None, results=None, tb=None, fb=None):
        self.col=col
        self.value=value
        self.results=results
        # FIXME: tb s/b true_branch
        self.tb=tb
        self.fb=fb


def build_tree(table, field=-1, scoref=entropy, ignore_fields=('pk', 'id')):
    """Build a classification decision tree

    >>> print_tree(build_tree(tobes_data))  # doctest: +NORMALIZE_WHITESPACE
    0:google? 
      T-> 3:21? 
          T-> {'Premium': 3}
          F-> 2:yes? 
              T-> {'Basic': 1}
              F-> {'None': 1}
      F-> 0:slashdot? 
          T-> {'None': 3}
          F-> 2:yes? 
              T-> {'Basic': 4}
              F-> 3:21? 
                  T-> {'Basic': 1}
                  F-> {'None': 3}
    """
    try:
        N = len(table)
    except:
        try:
            N = table.count()
        except:
            N = 0

    if not N:
        return DecisionNode()


    current_score=scoref(table)

    # Set up some variables to track the best criteria
    best_gain=0.0
    best_criteria=None
    best_sets=None
    if isinstance(table[0], dict) and isinstance(field, int):
        keys = sorted(tuple(k for k in table[0] if k not in ignore_fields))
        M = len(keys)
    else:
        M = len(table[0])
        keys = range(M)
    keyed_field = keys[field]

    for col in range(M):
        keyed_col = keys[col]

        if keyed_col == keyed_field:
            continue
        # Generate the list of different values in
        # this column
        column_values = set()
        for row in table:
            column_values.add(get(row, keyed_col))
        # Try dividing the table up for each value in this column
        for value in column_values:
            (set1, set2) = divide(table, field=col, target=value)
            set1, set2 = tuple(set1), tuple(set2)
            
            # Information improvement
            p = float(len(set1)) / N
            gain = current_score - p * scoref(set1) - (1 - p) * scoref(set2)
            if gain > best_gain and len(set1) > 0 and len(set2) > 0:
                best_gain = gain
                best_criteria = (col, value)
                best_sets = (set1, set2)

    # Create the sub branches   
    if best_gain > 0:
        trueBranch = build_tree(best_sets[0])
        falseBranch = build_tree(best_sets[1])
        return DecisionNode(col=best_criteria[0], value=best_criteria[1],
                            tb=trueBranch, fb=falseBranch)
    else:
        return DecisionNode(results=count_unique(table, field=keyed_field))  #keyed_field


def print_tree(tree, indent=' '):
    # if it's a leaf then results will be nonnull
    if tree.results is not None:
        print str(dict(tree.results))
    else:
        print str(tree.col) + ':' + str(tree.value) + '? '
        print indent + 'T->',
        print_tree(tree.tb,indent + '  ')
        print indent + 'F->',
        print_tree(tree.fb, indent + '  ')


def represent_tree(tree, indent=' '):
    s = ''
    # if it's a leaf then results will be nonnull
    if tree.results is not None:
        s += str(dict(tree.results)) + '\n'
    else:
        s += str(tree.col) + ':' + str(tree.value) + '?\n'
        s += indent + 'T->'
        s += represent_tree(tree.tb,indent + '  ') + '\n'
        s += indent + 'F->'
        s += represent_tree(tree.fb, indent + '  ') + '\n'
    return s


def get_width(tree):
    if tree.tb == None and tree.fb == None:
        return 1
    return get_width(tree.tb) + get_width(tree.fb)


def get_depth(tree):
    '''Walk a tree to get it's maximum depth (a nonnegative integer)'''
    if tree.tb == None and tree.fb == None:
        return 0
    if tree is None:
        return None
    return max(get_depth(tree.tb), get_depth(tree.fb)) + 1


def get(obj, key, default=None):
    '''Get a numbered element from a sequence or dict, or the attribute of an object'''
    if isinstance(obj, dict):
        if isinstance(key, int):
            try:
                return obj.get(sorted(tuple(obj))[key], default)
            except:
                pass
        try:
            return obj.get(key, default)
        except:
            pass
    try:
        # integer key and sequence (list) object
        return obj[key]
    except:
        return obj.getattr(key, default)

    








"""
from PuffineWareLLC.com whitepaper on LSA -- Puffinware makes a Windows Search tool
"""
from scipy.linalg import svd
#following needed for TFIDF
from math import log
import numpy as np


class PuffinLSA(object):

    def __init__(self, ignore_words=None, ignore_chars=None):
        self.ignore_words = ['and', 'or', 'for', 'in', 'on', 'of', 'near', 'to', 'a'] if ignore_words is None else ignore_words
        self.ignore_chars = r",:'!" if ignore_chars is None else ignore_chars
        self.wdict = {}
        self.keys = self.A = self.U = self.S = self.Vt = None
        self.dcount = 0        

    def parse(self, doc):
        words = doc.split()
        for w in words:
            w = w.lower().translate(None, self.ignore_chars)
            if w in self.ignore_words:
                continue
            elif w in self.wdict:
                self.wdict[w].append(self.dcount)
            else:
                self.wdict[w] = [self.dcount]
        self.dcount += 1      

    def build(self):
        self.keys = [k for k in self.wdict.keys() if len(self.wdict[k]) > 1]
        self.keys.sort()
        self.A = np.zeros([len(self.keys), self.dcount])
        for i, k in enumerate(self.keys):
            for d in self.wdict[k]:
                self.A[i,d] += 1

    def calc(self):
        self.U, self.S, self.Vt = svd(self.A)

    def TFIDF(self):
        WordsPerDoc = np.sum(self.A, axis=0)        
        DocsPerWord = np.sum(np.asarray(self.A > 0, 'i'), axis=1)
        rows, cols = self.A.shape
        for i in range(rows):
            for j in range(cols):
                self.A[i,j] = (self.A[i,j] / WordsPerDoc[j]) * log(float(cols) / DocsPerWord[i])

    def printA(self):
        print 'Here is the count matrix'
        print self.A

    def printSVD(self):
        print 'Here are the singular values'
        print self.S
        print 'Here are the first 3 columns of the U matrix'
        print -1*self.U[:, 0:3]
        print 'Here are the first 3 rows of the Vt matrix'
        print -1*self.Vt[0:3, :]




"""
Nonworking experiement with naive Bayes classification NLP

Reference:
    Bird, Klein, Loper, 2009 "Natural Language Processing with Python", O'REILLY
"""

from collections import defaultdict, Counter
from nlp.util import get_words


class Basic:
    def __init__(self, get_features=get_words, s='', path='', threshold=None):
        self.get_features = get_features
        # {category: {feature: num_occurences,...}}
        self.feature_count = defaultdict(Counter)
        # {category: total_num_occurrences}
        self.category_count = Counter()  # dict of Counters, count of documents binned in each category 
        self.threshold = defaultdict(lambda: 1.)
        if threshold:
            self.threshold.update(threshold)
        # self.num_items = 0

    def feature_probability(self, feature, category):
        return float(self.feature_count[category][feature]) / self.category_count[category]

    def weighted_feature_probability(self, feature, category, weight=1., assumed_probability=.5):
        """
        Use an assumed probability (for 0 occurence features) and weight it to produce a pseudo-probability
        >>> import pug.nlp.data.tobes as e
        >>> cl = Basic()
        >>> cl.train(e.training_set)
        >>> cl.weighted_feature_probability('quick', 'good')
        0.625
        >>> cl.weighted_feature_probability('money', 'good')
        0.25
        >>> cl.train(e.training_set)
        >>> cl.weighted_feature_probability('money', 'good')  # doctest: +ELLIPSIS
        0.1666...
        """

        total_feature_count = sum((self.feature_count[c][feature] for c in self.category_count))
        return  (weight * assumed_probability + total_feature_count * self.feature_probability(feature, category)) / (weight + total_feature_count)

    def increment_feature_count(self, feature, category):
        "Increase the count for a feature<->category association"
        # if isinstance(feature, (list, )):   # don't do tuple, or hasasstr(__iter__) in case feature = N-gram stored as tuple
        #     self.feature_count[category] += Counter(feature)
        # self.feature_count.setdefault(feature, Counter())
        # self.feature_count[feature].setdefault(category, 0)
        # self.feature_count[feature][category] += 1
        self.feature_count[category][feature] += 1

    def increment_category_count(self, category):
        "Increase the count of documents associated with a category"
        # self.category_count.setdefault(category, 0)
        self.category_count[category] += 1

    def count_feature_in_category(self, feature, category):
        "Count of feature occurences in a category"
        # using a defaultdict everywhere would take care of this
        #if category in self.feature_count and feature in self.feature_count[category]:
        return float(self.feature_count[category][feature])
        
    def items_in_category(self, category):
        if category in self.category_count:
            return float(self.category_count[category])    
        return 0.

    def num_items(self):
        "Total of the number of items that have been classified"
         # inefficient?, so cache this, if possible to detect with a dict has changed efficiently
        return sum(self.category_count.values())

    def num_categories(self):
        "Total of the number of items that have been classified"
         # inefficient?, so cache this, if possible to detect with a dict has changed efficiently
        return len(self.category_count)

    def categories(self):
        return list(self.category_count)

    def train(self, string, category=None):
        """Identify features (N-grams or words) in a string and associate that string with a category
        >>> c = Basic()
        >>> c.train('The quick brown fox jumps over the lazy dog', 'good')
        >>> c.train('Make quick money in the online casino', 'bad')
        >>> c.count_feature_in_category('quick', 'good')
        1.0
        >>> c.count_feature_in_category('quick', 'bad')
        1.0
        >>> c.num_items()
        2
        """
        if isinstance(string, basestring):
            self.feature_count[category] += Counter(self.get_features(string))
            self.category_count[category] += 1
        else:
            for cat, strings in string.iteritems():
                for s in strings:
                    self.train(s, cat)

    def __repr__(self):
        return repr(self.feature_count)

    def __str__(self):
        return str(self.feature_count)


class NaiveBayes(Basic):

    def repeated_item_probability(self, item, category):
        item_feature_count = Counter(get_words(item))
        p = 1.
        for f, c in item_feature_count.iteritems():
            # this doesn't make sense unless you're trying to compare the probabilities for numerouse vs single occurences of a feature
            p *= pow(self.weighted_feature_probability(f, category), c)
        return p
    
    def item_probability(self, item, category):
        item_features = set(get_words(item))
        p = 1.
        for f in item_features:
            p *= self.weighted_feature_probability(f, category)
        return p
    
    def category_probability(self, item, category):
        prior_category_probability = float(self.category_count[category]) / self.num_categories()
        return self.item_probability(item, category) * prior_category_probability

    def classify(self, item, default=None):
        max_prob = 0.
        next_prob = 0.
        best_match = default
        for category, count in self.category_count.iteritems():
            prob = self.item_probability(item, category)
            if prob > max_prob:
                next_prob = max_prob
                max_prob = prob
                if max_prob > next_prob * self.threshold:
                    best_match = category
        return best_match




# -*- coding: utf-8 -*-
""" By LiZhong Zheng, et. al"""
from __future__ import division

import os
from math import log, e
from collections import Mapping, Counter
import json

import nltk
import numpy as np

import character_subset as ascii
from util import is_ignorable_str, get_words


# trial for sharing code

IGNORE_FILES = ('readme',)

# john is editing from his github account

DOCUMENT_FOLDER_SUFFIX = 'pug/nlp/data'
for depth in range(7):
    if depth:
        DATA_FOLDER = os.path.join(os.path.sep.join(['..'] * depth), DOCUMENT_FOLDER_SUFFIX)
    else:
        DATA_FOLDER = DOCUMENT_FOLDER_SUFFIX
    DOCUMENT_FOLDER = os.path.join(DATA_FOLDER, 'inaugural_speeches')
    # print DOCUMENT_FOLDER
    try:
        DEFAULT_FILE_LIST = os.listdir(DOCUMENT_FOLDER)
        break
    except:
        DEFAULT_FILE_LIST = []
# print DATA_FOLDER


def get_adjacency_matrix(filenames=DEFAULT_FILE_LIST, entropy_threshold=0.90, folder=DOCUMENT_FOLDER, normalized=False, verbosity=1):
    """Calculate relevance score for words in a set of files

    score_words_in_files(filenames=DEFAULT_FILE_LIST, entropy_threshold=0.9):
        return scores, occurrence_matrix, filenames, most_relevant_words

    by LiZhong Zeng
    """

    stem = nltk.stem.PorterStemmer().stem
    filtered_filenames = []

    Text = []     # list of segmented texts
    Total = []    # all texts put together 
    Speech_Length = []
    
    if verbosity > 0:
        print 'Reading %s files' % len(filenames)
    i = 0
    for fn in filenames:
        if is_ignorable_str(fn, ignorable_strings=IGNORE_FILES, lower=True, filename=True, startswith=True):
            continue
        filtered_filenames += [fn]
        i += 1
        f = open(os.path.join(folder, fn))
        if verbosity > 1:
            print 'Reading ' + f.name
        raw = f.read()
        tokens = [stem(w) for w in nltk.word_tokenize(raw)]
        # delete short words and make everything lowercase
        tokens=[w.lower() for w in tokens if len(w) > 2]
        Speech_Length += [len(tokens)]
        Text.append(tokens)        
        Total=Total+tokens
    if verbosity > 0:
        print '%s files were indexed (%s were ignored)' % (i, len(filenames)-i)
            
    Empirical_Total=nltk.FreqDist(Total)    
    Vocabulary=Empirical_Total.keys()   # the entire set of words
    Size=len(Vocabulary)
    #numDoc=len(Text)

    Dist=range(Size)
    Vectors=[]          # Record a list of empirical distributions
    for i in range(len(filtered_filenames)):
        fdist=nltk.FreqDist(Text[i])

        for j in range(Size):
            Dist[j]=fdist[Vocabulary[j]]
            
        Vectors.append(Dist[:]) # Dist[:] makes a copy of the list to append
    
    Word_Relevance=range(Size) # store a relevance score for each word
    for wordIndex in range(Size):
        Word_Dist= nltk.FreqDist([Vectors[i][wordIndex] for i in range(len(filtered_filenames))])

        Word_Relevance[wordIndex] = 0

        # # check if the number of files that do not have the word is close to half
        # if (abs(Word_Dist[0] - len(filenames)/2) <= 3):
        #    Word_Relevance[wordIndex]=1            

        # information entropy normalized by the support (number of events)
        H_normalized = renyi_entropy(Word_Dist, alpha=.6, base=e, normalized=True)

        if (H_normalized > entropy_threshold):
            Word_Relevance[wordIndex] = 1

    Key_words = [Vocabulary[i] for i in range(Size) if Word_Relevance[i]] 

    if verbosity > 0:
        print 'Computed a relevance score for %s words and reduced it to %s words above %s%% relevance.' % (Size, len(Key_words), entropy_threshold * 100.)

    Reduced_Vectors=[]    
    for i in range(len(filtered_filenames)):
        total_count = sum(Vectors[i])
        if normalized: 
            Reduced_Vectors.append([Vectors[i][j] / total_count for j in range(Size) if Word_Relevance[j]])
        else:
            Reduced_Vectors.append([Vectors[i][j] for j in range(Size) if Word_Relevance[j]])
    return Reduced_Vectors, filtered_filenames, Key_words

def svd_scores(vectors, labels=None, verbosity=1):
    labels = labels or [str(i) for i in range(len(vectors))]

    if len(labels) != len(vectors):
        raise RuntimeWarning("The number of labels provided doesn't matc the number of vectors.")

    U, s, V = np.linalg.svd(vectors)

    scores = []

    if verbosity > 0:
        print 'SCORES', ':', 'LABELS'    
    for i, vector in enumerate(vectors):
        scores += [np.inner(V[0], vector) / sum(vector)]
        if verbosity > 0:
            print scores[-1], ':', labels[-1]
    return scores


def get_occurrence_matrix(strings, row_labels=None, tokenizer=get_words, stemmer=None):
    if stemmer is None:
        stemmer = get_occurrence_matrix.stemmer
    stemmer = stemmer or unicode
    print stemmer
    O_sparse = [Counter(stemmer(w) for w in tokenizer(s)) for s in strings]
    return matrix_from_counters(O_sparse, row_labels=row_labels)
get_occurrence_matrix.stemmer = nltk.stem.porter.PorterStemmer().stem


def sum_counters(counters):
    total = Counter()
    for counter in counters:
        total += counter
    return total


def matrix_from_counters(counters, row_labels=None):
    if not row_labels and isinstance(counters, Mapping) and all(isinstance(counter, Counter) for row_label, counter in counters.iteritems()):
        # TODO: def list_pair_from_dict(dict):
        row_labels = []
        new_counters = []
        for row_label, counter in counters.iteritems():
            row_labels += [row_label]
            new_counters += [counter]
        counters = new_counters
    if not row_labels or len(row_labels) != len(counters):
        row_labels = [i for i, c in enumerate(counters)]


    total = sum_counters(counters)

    col_labels, O = list(total), []
    for counts in counters:
        O += [[0] * len(col_labels)]
        for label, count in counts.iteritems():
            j = col_labels.index(label)
            O[-1][j] += count
    return O, row_labels, col_labels

def score_words_in_files(filenames=DEFAULT_FILE_LIST, entropy_threshold=0.90, folder=DOCUMENT_FOLDER, verbosity=1):
    adjacency_matrix, files, words = get_adjacency_matrix(filenames=DEFAULT_FILE_LIST, entropy_threshold=0.90, folder=DOCUMENT_FOLDER)
    return svd_scores(adjacency_matrix, files, verbosity), adjacency_matrix, files, words 


def reverse_dict(d):
    return dict((v, k) for (k, v) in dict(d).iteritems())


def reverse_dict_of_lists(d):
    ans = {}
    for (k, v) in dict(d).iteritems():
        for new_k in list(v):
            ans[new_k] = k
    return ans


def shannon_entropy(discrete_distribution, base=e, normalized=True):
    """Shannon entropy (information uncertainty) for the logarithm base (units of measure) indicated

    The default logarithm base is 2, which provides entropy in units of bits. Shannon used
    bases of 2, 10, and e (natural log), but scipy and most math packages assume e.

    >>> from scipy.stats import entropy as scipy_entropy
    >>> scipy_entropy([.5, .5]) / log(2) # doctest: +ELLIPSIS
    1.0
    >>> shannon_entropy([.5, .5], base=2)
    0.69314718...
    >>> scipy_entropy([.5, .5])  # doctest: +ELLIPSIS
    0.69314718...
    1.0
    >>> shannon_entropy([.5, .5], base=2, normalized=False)  # doctest: +ELLIPSIS
    0.69314718...
    >>> scipy_entropy([c / 13. for c in [1, 2, 3, 4, 3]])  # doctest: +ELLIPSIS
    1.5247073930...
    >>> shannon_entropy([1, 2, 3, 4, 3], normalized=False)
    1.5247073930...
    """
    if isinstance(discrete_distribution, Mapping):
        discrete_distribution = discrete_distribution.values()
    if base == None:
        base = e

    if not len(discrete_distribution):
        raise RuntimeWarning("Empty discrete_distribution probability distribution, so Renyi entropy (information) is zero.")
        return float('-0.')
    if not any(discrete_distribution):
        raise RuntimeWarning("Invalid value encountered in divison, 0/0 (zero divided by zero). Sum of discrete_distribution (discrete distribution integral) is zero.")
        return float('nan')
    if any(count < 0 for count in discrete_distribution):
        raise RuntimeWarning("Some counts or frequencies (probabilities) in discrete_distribution were negative.")

    total_count = float(sum(discrete_distribution))

    if base == e:
        H = sum(count * log(count / total_count) for count in discrete_distribution) / total_count
    else:
        H = sum(count * log(count / total_count, base) for count in discrete_distribution) / total_count
    if normalized:
        return -1 * H / log(len(discrete_distribution), base)
    return -1 * H


def renyi_entropy(word_counts, alpha=1, base=2, normalized=True):
    """Renyi entropy of order alpha and the logarithm base (units of measure) indicated

    The default logarithm base is 2, which provides entropy in units of bits, Renyi's 
    preferred units. Shannon used bases of 2, 10, and e (natural log).

    >>> from scipy.stats import entropy as scipy_entropy
    >>> scipy_entropy([.4, .6])  # doctest: +ELLIPSIS
    0.97...
    >>> renyi_entropy([.4, .6])  # doctest: +ELLIPSIS
    0.97...
    >>> renyi_entropy([.4, .6], .6)
    0.97...
    >>> renyi_entropy([], .6)
    -0.0
    >>> renyi_entropy([0.] * 10, .6)
    nan
    """
    if isinstance(word_counts, Mapping):
        word_counts = word_counts.values()
    if base == None:
        base = e

    N = len(word_counts)
    if not N:
        raise RuntimeWarning("Empty word_counts probability distribution, so Renyi entropy (information) is zero.")
        return float('-0.')
    if not any(word_counts):
        raise RuntimeWarning("Invalid value encountered in divison, 0/0 (zero divided by zero). Sum of word_counts (discrete distribution integral) is zero.")
        return float('nan')

    if alpha == 1:
        return shannon_entropy(word_counts, base=base, normalized=normalized)

    total_count = float(sum(word_counts))    
    entropy_unif = log(N, base)  # log of the support, don't have to turn to float
    sum_pow = sum(pow(count / total_count, alpha) for count in word_counts)
    
    # log(x) is 75% faster as log(x, math.e), according to timeit on python 2.7.5
    # so split into 2 cases, base e (None), and any other base
    if base == e:
        log_sum_pow = log(sum_pow)
    else:
        log_sum_pow = log(sum_pow, base)

    H = log_sum_pow / (1 - alpha)

    if normalized:
        return H / (entropy_unif or 1.)
    else:
        return H


def zheng_normalized_entropy(Word_Dist, alpha=.6):
    """Renyi entropy of order alpha and the logarithm base (units of measure) indicated

    The default logarithm base is 2, which provides entropy in units of bits, Renyi's 
    preferred units. Shannon used bases of 2, 10, and e (natural log).

    >>> from scipy.stats import entropy as scipy_entropy
    >>> scipy_entropy([.4, .6])  # doctest: +ELLIPSIS
    0.97...
    >>> zheng_normalized_entropy([.4, .6], alpha=1)  # doctest: +ELLIPSIS
    0.97...
    >>> word_freq = [c / 13. for c in [1, 2, 3, 4, 3]]
    >>> zheng_normalized_entropy(dict((k, v) for (k, v) in enumerate(word_freq)))  # doctest: +ELLIPSIS
    0.966162932302...
    """
    keys=Word_Dist.keys()
    Entropy_Unif= log(len(keys))  # log of the support, don't have to turn to float
    totalcount= float(sum([Word_Dist[w] for w in keys]))
    
    Entropy= -sum([ Word_Dist[w]/totalcount * log(Word_Dist[w]/totalcount) for w in keys])
    
    # Renyi entropy of order alpha
    #Entropy = log(sum ([pow(float(Word_Dist[w])/totalcount, alpha) for w in keys]))/(1-alpha)

    return float(Entropy) / Entropy_Unif


def co_adjacency(adjacency_matrix, row_names, col_names=None, bypass_col_names=True, normalized=True, verbosity=1):
    """Reduce a heterogenous adjacency matrix into a homogonous co-adjacency matrix

    coadjacency_matrix, names = co_adjacency(adjacency_matrix, row_names, col_names, bypass_col_names=True)
    """
    bypass_indx = int(not (int(bypass_col_names) % 2))
    names = (row_names, col_names or row_names)[bypass_indx]

    A = np.matrix(adjacency_matrix)
    if normalized:
        size = (float(len(adjacency_matrix[0])), float(len(adjacency_matrix)))
        A = A / size[int(bypass_indx)]
    if not bypass_indx:
        return (A * A.transpose()).tolist(), names
    return (A.transpose() * A).tolist(), names


def len_str_to_group(s, bin_min=None, bin_width=None):
    if bin_min is None:
        bin_min = len_str_to_group.min
    bin_width = bin_width or len_str_to_group.width or 1
    return int(round((len(s) - bin_min + 1) / float(bin_width)))
len_str_to_group.min = 3
len_str_to_group.width = 3


def yr_str_to_group(s, bin_min=None, bin_width=None, digits=None):
    if bin_min is None:
        bin_min = yr_str_to_group.min
    if digits is None:
        digits = yr_str_to_group.digits
    bin_width = bin_width or yr_str_to_group.width or 1
    return int(round((float(s[:digits]) - bin_min) / float(bin_width)))
yr_str_to_group.min = 1770
yr_str_to_group.width = 3
yr_str_to_group.digits = 4


def scale_exp(value, in_min, in_max, out_min=1, out_max=10, exp=1):
    """Rescale a scalar value to fit within out_min and out_max

    >>> scale_exp(150, in_min=100, in_max=200, out_min=0, out_max=1)
    .5
    >>> scale_exp(150, in_min=100, in_max=200, out_min=0, out_max=1, exp=2)
    .25
    """
    return np.power((out_max - out_min) * (value - in_min) / float(in_max - in_min), exp)


def matrix_scale_exp(A, out_min=0, out_max=1, exp=1):
    """Apply scape_exp to all elements of a matrix, auto-calculating in_min and in_max
    """
    A = np.matrix(A)
    return scale_exp(A, np.min(A), np.max(A), out_min, out_max, exp=exp).tolist()


def strip_path_ext_characters(s, strip_characters=ascii.ascii_nonword):
    return str(s).split('/')[-1].split('.')[0].strip(strip_characters)


def d3_graph(adjacency_matrix, row_names, col_names=None, str_to_group=len_str_to_group, str_to_name=strip_path_ext_characters, str_to_value=float, num_groups=7., directional=True):
    """Convert an adjacency matrix to a dict of nodes and links for d3 graph rendering

    row_names = [("name1", group_num), ("name2", group_num), ...]
    col_names = [("name1", group_num), ("name2", group_num), ...]

    Usually row_names and col_names are the same, but not necessarily.
    Group numbers should be an integer between 1 and the number of groupings
    of the nodes that you want to display.

    adjacency_matrix = [
        [edge00_value, edge01_value, edge02_value...],
        [edge10_value, edge11_value, edge12_value...],
        [edge20_value, edge21_value, edge22_value...],
        ...
        ]

    The output is a dictionary of lists of vertexes (called "nodes" in d3)
    and edges (called "links" in d3):

    {
        "nodes": [{"group": 1, "name": "Alpha"}, 
                  {"group": 1, "name": "Beta"}, 
                  {"group": 2, "name": "Gamma"}, ...
                 ],
        "links": [{"source": 1, "target": 0, "value": 1}, 
                  {"source": 2, "target": 0, "value": 8}, 
                  {"source": 3, "target": 0, "value": 10}, 
                 ]
    }
    """
    if col_names is None:
        col_names = row_names

    nodes, links = [], []

    print '-' * 10
    # get the nodes list first, from the row and column labels, even if not square
    for names in (row_names, col_names):
        for i, name_group in enumerate(names):
            if isinstance(name_group, basestring):
                name_group = (str_to_name(name_group), str_to_group(name_group))
            node = {"name": name_group[0], "group": name_group[1] or 1}
            print node
            if node not in nodes:
                nodes += [node]                

    for i, row in enumerate(adjacency_matrix):
        for j, value in enumerate(row):
            links += [{"source": i, "target": j, "value": str_to_value(value)}]
            if directional:
                links += [{"source": j, "target": i, "value": str_to_value(value)}]

    return {'nodes': nodes, 'links': links}


def d3_graph_occurrence(adjacency_matrix, row_names, col_names=None, str_to_group=len_str_to_group, str_to_value=float, verbosity=1, directional=True):
    """Convert an adjacency matrix to a dict of nodes and links for d3 graph rendering

    row_names = [("name1", group_num), ("name2", group_num), ...]
    col_names = [("name1", group_num), ("name2", group_num), ...]

    Usually row_names and col_names are the same, but not necessarily.
    Group numbers should be an integer between 1 and the number of groupings
    of the nodes that you want to display.

    adjacency_matrix = [
        [edge00_value, edge01_value, edge02_value...],
        [edge10_value, edge11_value, edge12_value...],
        [edge20_value, edge21_value, edge22_value...],
        ...
        ]

    The output is a dictionary of lists of vertexes (called "nodes" in d3)
    and edges (called "links" in d3):

    {
        "nodes": [{"group": 1, "name": "Alpha"}, 
                  {"group": 1, "name": "Beta"}, 
                  {"group": 2, "name": "Gamma"}, ...
                 ],
        "links": [{"source": 1, "target": 0, "value": 1}, 
                  {"source": 2, "target": 0, "value": 8}, 
                  {"source": 3, "target": 0, "value": 10}, 
                 ]
    }
    """
    if col_names is None:
        col_names = row_names

    nodes, links = [], []
    N, M = len(row_names), len(col_names)
    j = 0

    if verbosity > 1:
        print '-' * 10
    # get the nodes list first, from the row and column labels, even if not square
    for group_num, names in enumerate([row_names, col_names]):
        for i, name in enumerate(names):
            node = {"name": name, "group": group_num + 1}
            if verbosity > 1:
                print j, node
            nodes += [node]
            j += 1


    # get the edges next
    for i, row in enumerate(adjacency_matrix):
        for j, value in enumerate(row):
            links += [{"source": i, "target": N + j, "value": str_to_value(value)}]
            if directional:
                links += [{"source": j, "target": i, "value": str_to_value(value)}]

    return {'nodes': nodes, 'links': links}

def write_file_twice(js, fn, other_dir='../miner/static'):
    with open(fn, 'w') as f:
        json.dump(js, f, indent=2)
    other_path = os.path.join(other_dir, fn)
    if os.path.isdir(other_dir):
        with open(other_path, 'w') as f:
            json.dump(js, f, indent=2)

def president_party(name):
    surname = strip_path_ext_characters(name.split(' ')[-1], strip_characters=ascii.ascii_nonletter)
    return president_party.surname_party.get(surname, '')
president_party.party = reverse_dict_of_lists(json.load(open(os.path.join(DATA_FOLDER, 'president_political_parties.json'), 'r')))
president_party.surname_party = dict((name.split(' ')[-1], party) for (name, party) in president_party.party.iteritems())


def group_to_party(group):
    return group_to_party.parties[group or 0] or ""
group_to_party.parties = ["", "Whig", "Democratic", "Republican", "Democratic-Republican", "Federalist"]

def party_to_group(s):
    """
    >>> group_to_party(party_to_group("Republican"))
    'Republican'
    >>> group_to_party(party_to_group("Democratic"))
    'Democratic'
    group_to_party(party_to_group(president_party("whatever/2001-Bush")))
    'Republican'
    group_to_party(party_to_group(president_party("whatever/1980-Reagan")))
    'Republican'
    group_to_party(party_to_group(president_party("whatever/2012-Obama")))
    'Democratic'
    group_to_party(party_to_group(president_party("whatever/1776-Washington")) or 0)
    ''
    group_to_party(party_to_group(president_party("whatever/12345-Jefferson")) or 0)
    'Whig'
    """
    return party_to_group.parties.get(s)
party_to_group.parties = dict((p, i) for (i, p) in enumerate(group_to_party.parties))


def generate_word_cooccurrence(adjacency_matrix, files, words, num_groups=7., path=os.path.join('..','miner','static','word_cooccurrence.json'), normalized=False):
    print len(adjacency_matrix), len(adjacency_matrix[0])
    print sum(adjacency_matrix[0]), sum(adjacency_matrix[1])
    O, names = co_adjacency(adjacency_matrix, row_names=files, col_names=words, bypass_col_names=False, normalized=normalized)
    print len(O), len(O[0])
    print sum(O[0]), sum(O[1])
    #O = matrix_scale_exp(O, out_min=1 ** 2, out_max=31 ** 2, exp=.5)
    len_str_to_group.min = min(len(s) for s in names)
    len_str_to_group.width = float(max(len(s) for s in names) - len_str_to_group.min) / num_groups
    graph = d3_graph(O, names, str_to_group=len_str_to_group)

    print len(graph['nodes']), len(graph['links'])/float(len(graph['nodes']))
    print sum(e['value'] for e in graph['links'] if e['source'] == 1), sum(e['value'] for e in graph['links'] if e['source'] == 2)
    print '-'*50
    json.dump(graph, open(path, 'w'), indent=2)
    return graph


def generate_document_cooccurrence(adjacency_matrix, files, words, num_groups=7., path=os.path.join('..','miner','static','doc_cooccurrence.json'), normalized=False):
    O, names = co_adjacency(adjacency_matrix, row_names=files, col_names=words, bypass_col_names=True, normalized=normalized)
    #O = matrix_scale_exp(O, out_min=1 ** .66, out_max=31 ** .66, exp=1.5)
    graph = d3_graph(O, [(strip_path_ext_characters(n), party_to_group(president_party(n)) or 1) for n in names])
    json.dump(graph, open(path, 'w'), indent=2)
    return graph


def generate_occurrence(adjacency_matrix, files, words, path=os.path.join('..','miner','static','occurrence.json')):
    #O = matrix_scale_exp(O, out_min=1 ** .66, out_max=31 ** .66, exp=1.5)
    graph = d3_graph_occurrence(adjacency_matrix, [strip_path_ext_characters(n) for n in files], [strip_path_ext_characters(n) for n in words])
    json.dump(graph, open(path, 'w'), indent=2)
    return graph


def normalize_adjacency_matrix(adjacency_matrix, rowwise=True):
    am_normalized = []
    if rowwise:
        for row in adjacency_matrix:
            row_normalized, norm = [], sum(row)
            for value in row:
                row_normalized += [value / norm]
            am_normalized += [row_normalized]
    return am_normalized


def generate_json():
    adjacency_matrix, files, words = get_adjacency_matrix(sorted(DEFAULT_FILE_LIST), entropy_threshold=0.90, normalized=False, verbosity=2)
    print len(adjacency_matrix), len(adjacency_matrix[0])
    print sum(adjacency_matrix[0]), sum(adjacency_matrix[1])
    normalized_adjacency_matrix = normalize_adjacency_matrix(adjacency_matrix)
    print len(normalized_adjacency_matrix), len(normalized_adjacency_matrix[0])
    print sum(normalized_adjacency_matrix[0]), sum(normalized_adjacency_matrix[1])
    generate_document_cooccurrence(normalized_adjacency_matrix, files, words, num_groups=5, normalized=False)
    generate_word_cooccurrence(normalized_adjacency_matrix, files, words, num_groups=5, normalized=False)
    generate_occurrence(normalized_adjacency_matrix, files, words)


def generate_matrix(normalized=True, matrix_only=False, format='json', path=os.path.join('..','miner','static','adjacency_matrix.json')):
    adjacency_matrix, files, words = get_adjacency_matrix(sorted(DEFAULT_FILE_LIST), entropy_threshold=0.92, normalized=False, verbosity=2)
    if normalized:
        adjacency_matrix = normalize_adjacency_matrix(adjacency_matrix)
    format = path.split('.')[-1].strip()
    if format == 'json':
        if matrix_only:
            js = adjacency_matrix
        else:
            js = {
                'column_labels': words,
                'row_labels': files,
                'matrix': adjacency_matrix,
                }
    else:
        raise NotImplementedError("I only know how to write json files.")
    json.dump(js, open(path, 'w'))
    return adjacency_matrix, files, words

if __name__ == '__main__':
    """
    # This will produce a 14 x 14 matrix, 
    >>> scores, adjacency_matrix, files, words = score_words_in_files(DEFAULT_FILE_LIST[:14], entropy_threshold=0.99)
    >>> coadjacency, names = co_adjacency(adjacency_matrix, row_names=files, col_names=words, bypass_col_names=True)
    >>> len(coadjacence), len(coadjacency[0], len(names)
    (14, 14, 14)
    """
    #adjacency_matrix, files, words = get_adjacency_matrix(sorted(DEFAULT_FILE_LIST), entropy_threshold=0.92, normalized=False, verbosity=2)
    #scores = svd_scores(adjacency_matrix, labels=files, verbosity=2)
    #graph = d3_graph_occurrence(adjacency_matrix, files, words)
    # print json.dumps(graph, indent=4)



"""Multivariate Linear Regression"""
from pug.crawler.models import WikiItem as Item
import numpy as np

# a [10000 x num_predictors] matrix inversion is intractible, so don't dare try it
MAX_NUM_RECORDS = 100000


def predictor_values(predictors=None, filter_dict=None, exclude_dict=None, predicted='wikiitem.modified', max_num_records=10000):
    from collections import OrderedDict as OD
    if predictors is None:
        # values in the database retrieved for assignment to values in an array of data for the fit
        predictors = (
            # Name, Django ORM queryset record python expression
            ('modified', 'title'),
            ('crawled', 'modified'),
            )
    predictors = OD(predictors)
    # if filter_dict is None:
    #     # subset of the data to perform the multivariate linear regression on
    #     filter_dict = (('serial_number__isnot', 'TRUNCATED'),)
    # filter_dict = OD(filter_dict)

    qs = Item.objects
    if filter_dict:
        qs = qs.filter(**filter_dict)
    if exclude_dict:
        qs = qs.exclude(**exclude_dict)
    qs = qs.all()

    # predictor variable values, one row for each variable, a column for each record in the queryset (value of the predictor)
    X = []

    import math

    math_env = dict([(name, getattr(math, name, lambda x: x)) for name in math.__dict__ if not name.startswith('__')])
    blacklist = ['credits', 'del', 'delattr', 'dir', 'dreload', 'file', 'frozenset', 'get_ipython', 'make_option', 'os', 'sys', 'eval', 'globals', 'locals', 'open', 'exec', 'execfile']
    whitelist = ['sum', 'pow', 'float', 'int', 'floor', 'len', 'list', 'max', 'min', 'oct', 'ord', 'tuple', 'dict', 'str', 'unicode', 'type', 'isinstance', 'hex', 'ord', 'hash']
    global_env = dict([(name, __builtins__.get(name, lambda x: x)) for name in __builtins__ if name in whitelist and name not in blacklist and not name.startswith('__')])

    safe_env = {
        "locals": None,
        "globals":  None,
        "__name__": None,
        "__file__": None,
        "__builtins__": None,
        }

    safe_env.update(global_env)
    safe_env.update(math_env)

    N = min(max_num_records, qs.count(), MAX_NUM_RECORDS)
    y = []

    for record in qs[:N]:
        X += [[]]
        for name, expression in predictors.iteritems():
            safe_env['case'] = record
            X[-1] += [eval(expression, safe_env)]
        y += [eval(predicted, safe_env)]

    return X, y


def compute_fit(predictor_matrix, predicted_values):
    A = np.matrix(predictor_matrix)
    y = np.hstack(predicted_values)
    beta = np.linalg.lstsq(A, y)
    return beta


def demo():
    import nlp.plotter
    A, y = predictor_values()
    beta, residuals, rank, singular_values = compute_fit(A, y)
    nlp.plotter.regressionplot(A, y, (beta, residuals, rank, singular_values))



# -*- coding: utf-8 -*-
# by Lizhong Zheng <lizhong@MIT.EDU>

import numpy as np
# import os
import random
import math

def GaussianDataFile():
    
    f=open('GaussianData','w')
    
    f.write('# Gaussian vector of length k, each comes from X_i= sqrt(1-r^2) X_{i-1} + r W_i \n')
    f.write('# First line: n=number of samples, k=dimension per sample, Nt= first Nt values training \n')
    
    n = 1000
    k = 8
    r = 0.4
    s = np.sqrt(1-r*r)

    f.write(str(n) +', '+ str(k)+', ' + str(0) +'\n')
    
    for i in range(n):
        x=np.random.normal()
        f.write("%2.4f" %x)
        
        for j in range(k-1):
            x=s*x+ r*np.random.normal()
            f.write(', '+ "%2.4f" %x)
            
        f.write('\n')
    f.close()
    
def MarkovDataFile():
    
    f=open('MarkovData','w')
    f.write('# two Markov chains both with uniform steady state distributions\n')
    
# two transition matrices, each marked by integers, with LCM = 10

#  Take the convention W1[i] is the conditional distribution P_Y|X=i    
    
    W1= [[0.4,0.3,0.2,0.1],[0.1,0.4,0.3,0.2],[0.2,0.1,0.4,0.3],[0.3,0.2,0.1,0.4]]
    W2= [[0.2,0.3,0.3,0.2],[0.2,0.2,0.3,0.3],[0.3,0.2,0.2,0.3],[0.3,0.3,0.2,0.2]]
    
    f.write('# First line: n=number of samples, k=dimension per sample, Nt= first Nt values training \n')
    
    n=1000
    k=200
    Nt=1
    f.write(str(n) +', '+ str(k)+', ' + str(Nt) +'\n')
    
    for i in range(n):
        U=random.randint(0,1)
        if U==0:
            WW= W1
        else:
            WW= W2
            
    # write the lable   
        
        f.write(" %4d" %U)    
        
    # first sample
        x=random.randint(0,3)
        f.write(", %4d" %x)
        for j in range(k-1):
        
        # generate next sample a given distribution 
            x=np.random.choice(4, p=WW[x])
            f.write(", %4d" %x)
        
        f.write('\n')
    
    f.close()

        
    
def ReadDataFile(filename):
    
    f= open(filename,'r')
    
    str=f.readline()
    
 # lines starting with '#' are descriptions   
    while (str[0]=='#'):
        str=f.readline()
  
 #  first useful line gives the size of the array. 
       
    templist=str.split(", ")
    
    n= int(templist[0])
    k= int(templist[1])
    str=templist[2]    
    Nt= int(str[: str.find('\n')])
    
    Data=[]
    for i in range(n):
        str=f.readline()
        templist=str.split(", ")
        Data.append([float(x) for x in templist])
        
    return n,k,Nt, Data
        
        
def SimpleSVD(filename):

    n,k,Nt, Data = ReadDataFile(filename)

   # Simple PCA
    
    Data= np.array (Data)

# Empirical covariance matrix
    Kx= np.transpose(Data).dot(Data)/n

# Eigenvalues and Eigenvectors    
    Lambda, U= np.linalg.eig(Kx)

    print(Lambda)      
    
def RandomScore(Nt, Data):

# metric[i][j] is a map that takes (X_{n-1}, X_n)=(i,j) to a random value   
    metric=[range(4)]*4
    for i in range(4):
        for j in range(4):
            metric[i][j]=np.random.random()

    score = range(len(Data))        
    for i in range(len(Data)):
        score[i]=0
        
        for j in range(len(Data[0])-Nt-1):
            score[i]+= metric[int(Data[i][Nt+j])][int(Data[i][Nt+j+1])]
        
        score[i]=score[i]/(len(Data[0])-Nt)

    return score
    
def CheckClasses(score, Data):
    index0=[i for i in range(len(Data)) if Data[i][0]==0]
    index1=[i for i in range(len(Data)) if Data[i][0]==1]
    score0 =[score[i] for i in index0]
    score1 =[score[i] for i in index1]
    
    avg0=np.sum(score0)/len(score0)
    avg1=np.sum(score1)/len(score1)

    var0= np.sum([(score0[i]-avg0)*(score0[i]-avg0) for i in range(len(score0))])/len(score0)    
    var1= np.sum([(score1[i]-avg1)*(score1[i]-avg1) for i in range(len(score1))])/len(score1)  
    print ("Number of samples in class 0 =%d\n" %len(score0))
    print("Avg score for class 0 = %4.4f \n" %avg0)
    print("Variance for class 0 = %4.4f \n" %var0)
    print ("Number of samples in class 1 =%d\n" %len(score1))
    print("Avg score for class 1 = %4.4f \n" %avg1)
    print("Variance for class 1 = %4.4f \n" %var1)

def SVDScore(Nt, Data):
       
    W1= [[0.4,0.3,0.2,0.1],[0.1,0.4,0.3,0.2],[0.2,0.1,0.4,0.3],[0.3,0.2,0.1,0.4]]
    W2= [[0.2,0.3,0.3,0.2],[0.2,0.2,0.3,0.3],[0.3,0.2,0.2,0.3],[0.3,0.3,0.2,0.2]]

# V is the nominal transition matrix between X_{n-1} and X_n, can be learned 
# from empirical distributions
    
    V=W1
    dim=len(W1)
    
    for i in range(dim):
        for j in range(dim):
            V[i][j]=(W1[i][j]+W2[i][j])/2
            
# W is the noisy observation matrix, which is assumed to be the quternary 
#  symmetric channel          
    
    W= [[0.4,0.2,0.2,0.2],[0.2,0.4,0.2,0.2],[0.2,0.2,0.4,0.2],[0.2,0.2,0.2,0.4]]

    U,L,Vt= np.linalg.svd(W)
# the 4 vectors are Vt[0] = [1/2,1/2,1/2,1/2]
#                   Vt[1] = [0, 0, 1, -1] /sqrt(2)
#                   Vt[2] = [0, 2, -1, -1] /sqrt(6)
#                   Vt[3] = [3, -1, -1, -1]/sqrt(12)     
    
    Px12 = np.ones((dim, dim))       # stores the joint distribution of two samples
    dPx12 = np.ones((dim, dim))      # stores the difference between joint distribution and indpendent
    
    for i in range(dim):
        for j in range(dim):
            Px12[i][j] = W[i][j] /4.0
            dPx12[i][j] = Px12[i][j] - 1/16.0
    
    Py12 = np.ones((dim, dim))    # stores the joint distribution of two noisy samples

    for i in range(dim):
        
        for j in range(dim):
            
            sum=0
            for ii in range(dim):
                for jj in range(dim):
                    sum+= Px12[ii][jj]*W[ii][i]*W[jj][j]
            
            Py12[i][j] = sum
   
# Generate the two symbol B matrix
    
    B= np.ones((dim*dim, dim*dim))
    
    for y1 in range(dim):
        for y2 in range(dim):
            for x1 in range(dim):
                for x2 in range(dim):
                    B[y1*dim+y2][x1*dim+x2]= W[x1][y1]*W[x2][y2]* np.sqrt(Px12[x1][x2]/Py12[y1][y2])
            
    
    Ub, Lb, Vb = np.linalg.svd(B)
    
#Lb[0]=1
    
    metric = Vb[1].reshape((dim, dim))
    
    score = range(len(Data))        
    for i in range(len(Data)):
        
        score[i]=0

        for j in range(len(Data[0])-Nt-1):
            
            score[i]+= metric[int(Data[i][Nt+j])][int(Data[i][Nt+j+1])]
        score[i]=score[i]/(len(Data[0])-Nt)

    return(score)   
    
def OptimalScore(Nt, Data, dim=4):

# The best score function
    W1= [[0.4,0.3,0.2,0.1],[0.1,0.4,0.3,0.2],[0.2,0.1,0.4,0.3],[0.3,0.2,0.1,0.4]]
    W2= [[0.2,0.3,0.3,0.2],[0.2,0.2,0.3,0.3],[0.3,0.2,0.2,0.3],[0.3,0.3,0.2,0.2]]
    
    metric = np.ones((dim, dim)) 
    
    for i in range(dim):
        for j in range(dim):
            metric[i][j]= math.log(W2[i][j]/W1[i][j])

    
    score = range(len(Data))        
    for i in range(len(Data)):
        
        score[i]=0

        for j in range(len(Data[0])-Nt-1):
            
            score[i]+= metric[int(Data[i][Nt+j])][int(Data[i][Nt+j+1])]
        score[i]=score[i]/(len(Data[0])-Nt)
   
   
    return(score)

def Top3Score(Nt, Data, dim=4):
    
    NTop = 6  # the number of top scores to be calculated
    
    W1= [[0.4,0.3,0.2,0.1],[0.1,0.4,0.3,0.2],[0.2,0.1,0.4,0.3],[0.3,0.2,0.1,0.4]]
    W2= [[0.2,0.3,0.3,0.2],[0.2,0.2,0.3,0.3],[0.3,0.2,0.2,0.3],[0.3,0.3,0.2,0.2]]

# V is the nominal transition matrix between X_{n-1} and X_n, can be learned 
# from empirical distributions
    
    V=W1
    dim=len(W1)
    
    for i in range(dim):
        for j in range(dim):
            V[i][j]=(W1[i][j]+W2[i][j])/2
            
# W is the noisy observation matrix, which is assumed to be the quternary 
#  symmetric channel          
    
    W= [[0.4,0.2,0.2,0.2],[0.2,0.4,0.2,0.2],[0.2,0.2,0.4,0.2],[0.2,0.2,0.2,0.4]]

    U,L,Vt= np.linalg.svd(W)
# the 4 vectors are Vt[0] = [1/2,1/2,1/2,1/2]
#                   Vt[1] = [0, 0, 1, -1] /sqrt(2)
#                   Vt[2] = [0, 2, -1, -1] /sqrt(6)
#                   Vt[3] = [3, -1, -1, -1]/sqrt(12)     
    
    Px12 = np.ones((dim, dim))       # stores the joint distribution of two samples
    dPx12 = np.ones((dim, dim))      # stores the difference between joint distribution and indpendent
    
    for i in range(dim):
        for j in range(dim):
            Px12[i][j] = W[i][j] /4.0
            dPx12[i][j] = Px12[i][j] - 1/16.0
    
    Py12 = np.ones((dim, dim))    # stores the joint distribution of two noisy samples

    for i in range(dim):
        
        for j in range(dim):
            
            sum=0
            for ii in range(dim):
                for jj in range(dim):
                    sum+= Px12[ii][jj]*W[ii][i]*W[jj][j]
            
            Py12[i][j] = sum
   
# Generate the two symbol B matrix
    
    B= np.ones((dim*dim, dim*dim))
    
    for y1 in range(dim):
        for y2 in range(dim):
            for x1 in range(dim):
                for x2 in range(dim):
                    B[y1*dim+y2][x1*dim+x2]= W[x1][y1]*W[x2][y2]* np.sqrt(Px12[x1][x2]/Py12[y1][y2])
            
    
    Ub, Lb, Vb = np.linalg.svd(B)
    
#Lb[0]=1
    
    scorearray=np.ones((NTop, len(Data)))
    score = range(len(Data))  
    
    for nTop in range(NTop):
        metric = Vb[1+nTop].reshape((dim, dim))
    
        for i in range(len(Data)):
        
            scorearray[nTop][i]=0

            for j in range(len(Data[0])-Nt-1):
            
                scorearray[nTop][i]+= metric[int(Data[i][Nt+j])][int(Data[i][Nt+j+1])]
            
            scorearray[nTop][i]=scorearray[nTop][i]/(len(Data[0])-Nt)

    S=np.matrix(scorearray)*np.matrix(scorearray).transpose() #covariance matrix
    Ls, Us = np.linalg.eig(S)  # Us[:,i] is the eigenvector for Ls[i]
    
    for i in range(len(Data)):
        score[i]= np.sum([Us[j,0]*scorearray[j,i] for j in range(NTop)])
        # take inner product to the first eigenvector 

    return(score)


def test():
      
    dim = 4
    # write 'GaussianData' to store a Gaussian Markov data file
    #    GaussianDataFile()
    #    SimpleSVD('GaussianData')   

    # generage mixed Markov data, uncomment to generate a new one
    # comment out for repeating on the same data
    MarkovDataFile()

    # read data in any data file    
    [n,k, Nt, Data] = ReadDataFile('MarkovData')

    # randomly pick a score function based on two adjacent samples    
    #    score =RandomScore(Nt, Data)
    #    CheckClasses(score, Data)

 
    # Compute SVD w.r.t. a noisy observation and score with it
    score = SVDScore(Nt, Data)
    print(" SVD Scores:\n ")
    CheckClasses(score, Data)

    # Compute the best metric from log likelihood, and score with it
    score = OptimalScore(Nt, Data, dim=dim)    
    print(" Optimal Scores: \n")
    CheckClasses(score, Data)
    
    # Compute the top 3 scores from SVD and then do a 3-D PCA
    score = Top3Score(Nt, Data)
    print(" Top 3 Score PCA :\n")
    CheckClasses(score, Data)
   

    Alpha = np.ones((dim, dim))   # Alpha is the coefficients of dPx12 w.r.t. Vt[i]\otimes Vt[j]
    
    for i in range(dim):

        for j in range(dim):

            sum=0
            
            for ii in range(dim):
                for jj in range(dim):
#                   sum+= dPx12[ii][jj] * Vt[i][ii]*Vt[j][jj]
                   sum+= i                    
                    
            Alpha[i][j] = sum

# Alpha[0][0] is useless, and from dPx12 to Alpha is a rotation, the following
# checks the normal is unchanged. 
#    sum1=0
#    sum2=0
#    for i in range(4):
#        for j in range(4):
#            sum1+=Alpha[i][j]**2
#            sum2+= dPx12[i][j]**2


if __name__ == "__main__":
    test()


#test_decider.py


from call_center.models import CaseExchange, CaseHDTVHeader, CaseMaster

from pug.db.explore import count_unique
from pug.nlp.db_decision_tree import build_tree, print_tree


N = CaseMaster.objects.count()
UN = CaseMaster.objects.values('case_number').distinct().count()
N_ce = CaseExchange.objects.count()
UN_ce = CaseExchange.objects.values('case_number').distinct().count()
N_hdtv = CaseHDTVHeader.objects.count()
UN_hdtv = CaseHDTVHeader.objects.values('case_number').distinct().count()


un = count_unique(CaseExchange.objects.values('case_number'), 'case_number')
assert(len(un.values()) == UN_ce)
assert(sum(un.values()) == N_ce)


qs = CaseHDTVHeader.objects.filter(case_number__lt=4000000)
ex = qs.all()[0]
ex.service_calls
print_tree(build_tree(qs, field='service_calls', ignore_fields=('id', 'case_number')))
# dispatch_status:Completed ? 
#  T-> date_time:2008-09-15 12:25:34.270000? 
#    T-> {1: 1}
#    F-> date_time:2008-07-09 08:49:36.437000? 
#      T-> {0: 1}
#      F-> {None: 0}
#  F-> {None: 0}
qs = CaseHDTVHeader.objects.filter(case_number__lt=2000000)
ex = qs.all()[0]
ex.service_calls
print_tree(build_tree(qs, field='dispatch_status', ignore_fields=('id', 'case_number', 'service_calls')))
# dispatch_status:Completed ? 
#  T-> date_time:2008-09-15 12:25:34.270000? 
#    T-> {1: 1}
#    F-> date_time:2008-07-09 08:49:36.437000? 
#      T-> {0: 1}
#      F-> {None: 0}
#  F-> {None: 0}

print_tree(build_tree(qs, field='dispatch_status', ignore_fields=('id', 'dispatch_rejected', 'case_number', 'service_calls', 'dispatch_reject_reason', 'dispatch_reject_reason', 'dispatch_create', 'dispatch_completed', 'dispatch_expired', 'dispatch_scheduled', 'dispatch_received')))
print_tree(build_tree(qs, field='dispatch_status', include_fields=('user_type', 'purchase_where', 'repair_exch', 'case_number__state', 'case_number__model', 'case_number__contact_method')))
print_tree(build_tree(qs, field='service_calls', ignore_fields=('id', 'case_number')))
# s_state:TX? 
#  T-> servicer_refno:4656388                  ? 
#    T-> date_time:2003-02-06 17:27:25.390000? 
#      T-> {3: 1}
#      F-> {2: 1}
#    F-> case_number:797996? 
#      T-> {None: 0}
#      F-> {1: 3}
#  F-> store_name:Cardinal TV and Audio              ? 
#    T-> date_time:2002-04-08 14:35:25.900000? 
#      T-> {2: 1}
#      F-> {1: 1}
#    F-> {None: 0}

query_set = CaseHDTVHeader.objects.filter(case_number__gt=3000000, case_number__lt=3500000)
tree = build_tree(query_set, field='dispatch_status', include_fields=(
    'user_type',
    'dispatch_status',
    'case_number__state',
    'case_number__model',
    'case_number__country',
    'case_number__zip',
    'case_number__queue_time',
    'case_number__problem',
    'case_number__call_type',
    'case_number__contact_method'))
print_tree(tree)

#print_tree(tree)

# case_number__queue_time:2008-01-29 10:44:12.537000? 
#  T-> case_number__queue_time:2008-01-29 12:01:47.393000? 
#    T-> case_number__queue_time:2009-03-23 15:21:10.390000? 
#      T-> case_number__problem:7E11? 
#        T-> {None: 0}
#        F-> case_number__problem:7D00? 
#          T-> case_number__state:TN? 
#            T-> {u'Completed ': 1, None: 0}
#            F-> {None: 0}
#          F-> case_number__model:LC42D62U         ? 
#            T-> {None: 0}
#            F-> case_number__state:NJ? 
#              T-> {u'Accepted  ': 1}
#              F-> case_number__state:OH? 
#                T-> {u'Accepted  ': 1}
#                F-> case_number__state:CA? 
#                  T-> {u'Accepted  ': 1}
#                  F-> case_number__state:HI? 
#                    T-> {None: 0}
#                    F-> case_number__model:KB6015KS         ? 
#                      T-> {None: 0}
#                      F-> user_type:None? 
#                        T-> {u'Completed ': 1}
#                        F-> case_number__state:MI? 
#                          T-> {u'Completed ': 1}
#                          F-> {u'Completed ': 2, None: 0}
#      F-> case_number__problem:7B00? 
#        T-> {u'Accepted  ': 1}
#        F-> case_number__model:LC52D82U         ? 
#          T-> {u'Completed ': 1}
#          F-> case_number__model:LC40C32U         ? 
#            T-> case_number__state:NY? 
#              T-> {u'Completed ': 1}
#              F-> {None: 0}
#            F-> case_number__contact_method:6? 
#              T-> {None: 0}
#              F-> case_number__problem:5011? 
#                T-> {u'Completed ': 1, None: 0}
#                F-> case_number__queue_time:2008-10-29 11:28:45.207000? 
#                  T-> case_number__queue_time:2008-10-29 18:47:01.450000? 
#                    T-> case_number__state:NY? 
#                      T-> {u'Completed ': 1, None: 0}
#                      F-> {None: 0}
#                    F-> {u'Completed ': 1}
#                  F-> {None: 0}
#    F-> {u'Cancelled ': 1, u'Accepted  ': 1}
#  F-> case_number__state:FL? 
#    T-> case_number__model:LC37D6U          ? 
#      T-> {u'Completed ': 1}
#      F-> case_number__model:LC26D4U          ? 
#        T-> {u'Completed ': 1}
#        F-> {None: 0}
#    F-> {None: 0}


from call_center.models import CaseHDTVHeader
from pug.nlp.db_decision_tree import build_tree, draw_tree, print_tree

tree_list = []
fields = ['dispatch_status']
qs_kwargs_list = [
    {'case_number__gt':3000000, 'case_number__lt': 3200000},
    {'case_number__gt':3200000, 'case_number__lt': 3600000},
    {'case_number__gt':3600000, 'case_number__lt': 4000000}
    ]
include_fields_list = [
    list('user_type', 'case_number__state', 'case_number__problem'),
    list('user_type', 'case_number__problem', 'case_number__model_number'),
    list('user_type', 'case_number__call_type', 'case_number__contact_method'),
    ]
for i, field in enumerate(fields):
    for j, qs_kwargs in enumerate(qs_kwargs_list):
        for k, include_fields in enumerate(include_fields_list):
            tree = [field, qs_kwargs, include_fields + field, build_tree(CaseHDTVHeader.objects.filter(**qs_kwargs), field=field, include_fields=include_fields)]
            tree_list += tree
            print
            print '=' * 80
            print field, qs_kwargs_list
            print include_fields
            print '-' * 80
            print_tree(tree)
            print '-' * 80
            draw_tree('tree_%s_%s_%s.jpg' % (i, j, k))